require 'class_extensions'

# toto je zde jen proto aby se zkrátila serializace a vešla do pole mysql. Konkrétně aby se nevuáděl pořád dokola typ objektu
class ArrayWithMySerialize < Array
  def self.dump(osm_relation_members)
    osm_relation_members ? osm_relation_members.collect{|hash| ActiveSupport::JSON.encode(hash) }.join("\n") : nil
  end
  def self.load(json)
    json ? json.split("\n").collect{ |string| ActiveSupport::JSON.decode(string) } : []
  end
end

# Objekt reprezentuje trail z OSM. Při importu se předzpracují některé další údaje jako
# gpx_segments - seznam koordinátů jednotlivých segmentů seskupených v pořadí dle trasy
class OsmTrail < ActiveRecord::Base
  has_many :osm_signs, -> { order(:km) }, :dependent => :destroy
  has_many :osm_guideposts, :through => :osm_signs
  has_one  :trail, :primary_key => :osmid, :foreign_key=> :osmid
  has_one  :note, primary_key: :osmid, foreign_key: :osmid, class_name: :OsmTrailNote, autosave: true
  scope :search, lambda { |term| where("CONCAT_WS('', kctref, LOWER(nazev)) LIKE ?", "%#{term.mb_chars.downcase
}%")  }

  serialize :osmtags
  serialize :osm_relation_members, ArrayWithMySerialize
  serialize :gpx_segments
  serialize :gpx_excursions
  
  delegate :point_with_kilometraz, :kilometraz_of_nearest_point_to, :length_without_gaps, to: :gpx
  delegate :error, to: :note, allow_nil: true
  include TrailCommons
  
  # distance of point to main trail
  def distance_to(point_or_osm_guidepost)
    point_or_osm_guidepost = point_or_osm_guidepost.coordinates if point_or_osm_guidepost.is_a? OsmGuidepost
    gpx.distance_to(point_or_osm_guidepost)
  end
  
  # distance of point to main trail or excursions
  def distance_to_including_excursions(point_or_osm_guidepost)
    point_or_osm_guidepost = point_or_osm_guidepost.coordinates if point_or_osm_guidepost.is_a? OsmGuidepost
    excursions = GpxWithGaps.new(self.gpx_excursions)
    dist_to_excursions = excursions.distance_to(point_or_osm_guidepost)
    return [ gpx.distance_to(point_or_osm_guidepost), dist_to_excursions ].compact.min
  end
  
  # create or update from json - tries to preserve ids and relations if they exist
  def self.create_or_update_from_json(json)
    logger.error "TOOLS> create_or_update_from_json started"

    result = OsmTrail.new({
      :kctref  => json['tags']['ref'] || json['tags']['ref:kct'],
      :osmid   => json['id'],
      :nazev   => json['tags']['name'],
      :osmtags    => json['tags'].reject{|k,v| ['type#route','route#hiking'].include?(k+'#'+v) },
      :osm_relation_members => json['members']
    })
    
    result.nazev_normalized = result.nazev.normalize unless result.nazev.nil?
    result.barva = result.osmbarva
    result.typ = result.osmtyp

    # destroy previous copy of the trail
    existing = OsmTrail.find_by_osmid(result.osmid)
    if !existing.nil?
      existing.destroy
    end
    
    result.save   # otherwise members can not be connected
   
    # read node and way members of relation and connect guideposts
    errors = []
    ways = []
    result.number_of_ends = 0
    
    puts "\nImporting from json (#{json['id']})"
    json['members'].each do |member_hash|
      case member_hash['type']
        when 'node'
          print 'n'
          raise "Rozcestník bez osmid: #{member_hash.inspect}" if member_hash['ref'].nil?
          guidepost = OsmGuidepost.find_by_osmid(member_hash['ref'])
          if guidepost.nil?  
            if member_hash['role']=='guidepost'
              errors << "Na trase #{result.inspect_ref}  je rozcestník #{member_hash['ref']}, který není v databázi."  
            end
          else   
            result.osm_guideposts << guidepost if !result.osm_guideposts.include?(guidepost)
          end
        when 'way'
          print 'w'
          if member_hash['role']!='stop'
            way = OsmWay.find(member_hash['ref'])
            errors << "Nenalezena cesta s osmid #{member_hash['ref']}. Cesty z relací by měly být v souboru." if way.nil?
            ways << way unless way.nil?
          end
          if member_hash['role']=='end' or member_hash['role']=='start'
            result.number_of_ends += 1 
          end
        when 'relation'
          print 'r'
          # nic - tohle mohou být podtrasy a s tím zatím neumíme 
        else
          print '?'
          errors << "Člen neznámého typu '#{member_hash['type']}'v relaci trasy."
        end
    end
    
    print '.'
    result.initialize_gps_segments_from_ways(ways)
    print '.'
    result.initialize_kilometraz
    print '.'
    
    result.save # may not be needed. I do not know.
    logger.error errors.collect{|mess| "TOOLS> #{mess}" }.join("\n")
    return result
  end
  
  def priority; note.andand.priority end
  def priority=(value)
    self.note = OsmTrailNote.new() if note.nil?
    self.note.priority = value
  end
  
  # k signs doplni doplni sloupec km (kilometraz)
  def initialize_kilometraz
    # beware - this needs that gpx_segments are already sorted
    self.osm_signs.each do |os|
      os.km = os.km_on_trail
      os.save
    end
  end
  
  def initialize_gps_segments_from_ways(ways)
    # vytvoření segmentů
    number_of_ends = 0
    segments = ways.collect{ |way| way.coordinates_array }
    begin # postupně přeskupuje a propojuje segmenty # TODO: identifikovat odbočky
      something_changed = false
      (segments.length-1).downto(0) do |i|
        (i-1).downto(0) do |j|
          if  segments[j].first  == segments[i].first
            segments[j][0..0] = segments[i].reverse    # obrať a přidej na začátek
          elsif segments[j].last == segments[i].first
            segments[j][-1..-1] = segments[i]          # přidej na konec
          elsif segments[j].first== segments[i].last
            segments[j][0..0]  = segments[i]           # přidej na začátek
          elsif segments[j].last == segments[i].last
            segments[j][-1..-1] = segments[i].reverse  # obrať přidej na konec
          else
            next    # nesedí, jedeme dál
          end
          # jsem-li tady něco se povedlo
          segments.delete_at(i)
          something_changed = true
          break
        end
      end
    end while something_changed   # až dokud už není co propojit
    
    # v tento moment jsou segments pospojované tak, že už je nelze více pospojovávat. Nyní musíme identifikovat odbočky
    # budeme naivisticky předpokládat jediný případ a to kdy je nějaká odbočka napojena vprostřed jiného segmentu
    # odbočky přitom vyházíme do gpx_excursions, zatím vždy tu ze tří možností, která je nejgratší
    self.gpx_excursions = []
    GpxWithGaps.new(segments).branching_points.each do |bp| # trochu o šajstli, protože v cyklu pole modifikuji, ale kvůli nextům to nejde
      logger.info "BP #{bp}"
      
      # do segpom si nejdříve dáme všechny segmenty které obsahují bp
      segpom = segments.select{|seg| seg.include?(bp)}
      next if segpom.size!=2 # TODO:issue#2 
      ## nyní chceme do seg1 dát segment, který už má bp na kraji a do segP ten který ho má vprostřed
      case bp
        when segpom[0].first then seg1 = segpom[0]; segP = segpom[1]
        when segpom[0].last  then seg1 = segpom[0].reverse; segP = segpom[1]
        when segpom[1].first then seg1 = segpom[1]; segP = segpom[0]
        when segpom[1].last  then seg1 = segpom[1].reverse; segP = segpom[0]
        else next # TODO:issue#2
      end
       
      ## vyrobíme si trojčlenné pole tak, že první bod všech členů je bp (u seg1 už to je zařízený
      segtroj = [seg1, segP.slice(segP.index(bp)..-1), segP.slice(0..segP.index(bp)).reverse]
      
      ## zjistíme, který člen je nejkratší
      segtroj_sorted_with_lengths = segtroj.collect{ |seg| {length: GpxSegment.new(seg).length, seg: seg}}.sort_by{ |hash| hash[:length] }
      
      ## vložíme nové členy na příslušná místa (předtím vyndáme původní členy z pole)
      joinback = segtroj_sorted_with_lengths[1][:seg].reverse # otočené pole 
      joinback.pop # odebrat poslední prvek aby se neopakoval (je to bp)
      joinback += segtroj_sorted_with_lengths[2][:seg]
      segments.delete(segpom[0]); segments.delete(segpom[1]);
      segments << joinback
      self.gpx_excursions << segtroj_sorted_with_lengths[0][:seg]
      
    end
    
    # assign and reverse if needed
    self.gpx_segments = segments
    self.set_apropriate_order_of_gpx_segments
    
    # recalculate delka
    self.delka = self.length_without_gaps
    self.save
  end    
  
  def set_apropriate_order_of_gpx_segments
    gpx_with_gaps = self.gpx
    gpx_with_gaps.sort!
    gpx_with_gaps.reverse! if reversed?
    self.gpx_segments = gpx_with_gaps.coordinates_array
  end
  
  def reverse!
    self.reversed = !self.reversed?
    self.set_apropriate_order_of_gpx_segments
    self.initialize_kilometraz
    self
  end
  
  def sorted; self end # sorted on inmport
  
  def osm_signs_sorted
    osm_signs # .to_a.collect { |os| [os.km,os] }.sort {|a,b| a[0]<=>b[0]}.collect{|a| a[1]} # sortuje se to při nahrávání
  end
  
  def osm_guideposts_sorted; osm_signs_sorted.collect { |os| os.osm_guidepost} end
  def self.all_by_trail(trail);  OsmTrail.where(:kctref => trail.kctref)   end
  
  def osmbarva # najdu barvu podle tagů a zkontroluji jestli jsou konsistentní
    # najde barvu podle osmc:symbol
    osmcsymbol_barva = nil
    begin
      osmcsymbol_barva = OsmTrail.color2barva(osmtags['osmc:symbol'].split(':').first) if osmtags['osmc:symbol'] 
    rescue StandardError => e; add_error "Chyba při rozpoznávání barvy: #{e.message}" end
    add_error "Podezřelá OSM barva #{osmcsymbol_barva}" unless osmcsymbol_barva.nil? or osmcsymbol_barva.match(/^červená|modrá|zelená|žlutá$/)
    
    # najde barvu podle tagu kct_barva
    kct_x_barva =  osmtags.keys.grep(/kct_(yellow|green|blue|red)/){OsmTrail.color2barva($1)}
    add_error "Conflicting kct_* tags found" if kct_x_barva.size>1
    kct_x_barva = kct_x_barva.first

    # srovnám nalezená a vydám výsledek
    if !osmcsymbol_barva.nil? and !kct_x_barva.nil? and kct_x_barva!=osmcsymbol_barva
      add_error "Conflicting colour tags found: #{kct_x_barva}, #{osmcsymbol_barva}"
    end
    return osmcsymbol_barva || kct_x_barva
  end
  
  def osmtyp # najdu typ podle tagů
    begin
      !osmtags.nil? and !osmtags['osmc:symbol'].nil? ?
        OsmTrail.osmcsymbol2typ(osmtags['osmc:symbol']) : nil
#     rescue
#       raise "Unrecognized osmcsymbol #{osmtags['osmc:symbol']} for #{self.inspect}"
    end
  end
  #def typ; osmtyp end # for testing

  def inspect_ref
    kctref || osmid
  end
    
  def inspect_nazev
    return nazev unless nazev.blank?
    return "Turistická značená trasa KČT #{kctref}" unless kctref.blank?
    return "Značená trasa OSM #{osmid}" unless osmid.blank?
    'N/A'
  end

  def kct_trail
    trail.andand.kct_trail
  end

  def kct_trail_estimate
    kct_trail || ( kctref.nil? ? nil : KctTrail.find_by_kctref(self.kctref) )
  end
  
  def kct_trail_candidates_hash
    kct_trail_hash = Hash.new
    self.osm_guideposts.select{|og| og.exists?}.each do |osm_guidepost| 
      kct_trails = osm_guidepost.kct_guidepost.andand.kct_trails.andand.where(:barva=>self.barva) || []
      kct_trails.each do |kct_trail|
        kct_trail_hash[kct_trail] = [] if kct_trail_hash[kct_trail].nil?
        kct_trail_hash[kct_trail] << osm_guidepost
      end
    end
    kct_trail_hash
  end
  
  def trail
    Trail.find_by_osmid(self.osmid)
  end
  
  def importable?
    trail.nil? and !kct_trail.nil?
  end
  
  def to_s
    "osm:#{osmid}"
  end
  
  def add_error(aerror)
    if error.nil? 
      error=aerror
    else
      error+="\n"+aerror
    end
  end
  
  def gaps
    gpx.gaps
  end
  
  def gpx
    @gpx ||= GpxWithGaps.new(self.gpx_segments)
  end
  
  def trail_relative_distance(sign1,sign2); sign2.km - sign1.km end
  def trail_distance(sign1,sign2); trail_relative_distance(sign1,sign2).abs end
  
  def length_with_gaps; gpx.length_with_gaps end
  def length_with_gaps_string; Map.to_length_s(gpx.length_with_gaps) end
  
  # nalezne kandidátské rozcestníky blízko dané pozici, který vůbec není na trail ale někde v terénu
  def find_guideposts_near_position(kms)
    centre = gpx.point_with_kilometraz(kms)
    lat = centre[0]; lon = centre[1]
    result = OsmGuidepost.where("#{lat-0.009} < lat AND lat < #{lat+0.009} AND #{lon-0.0136} < lon AND lon < #{lon+0.0136} ") # prvostni nastrel
    result = result.select { |osm_guidepost| self.distance_to(osm_guidepost.coordinates)< 150 }  # vyfiltrova ty co jsou  blizko trasy
    result
  end
  
  def length
    gpx.length_with_gaps / 1000
  end
  
  def reversed?; note.andand.reversed == 1 end
  def reversed=(value)
    self.create_note if note.nil?
    if value
      note.reversed = 1
    else
      note.reversed = 0
    end
    note.save!
    self.set_apropriate_order_of_gpx_segments
    self.save
  end
  def exists?; !trail.nil? end
  def error?;  !error.nil? end
  def number_of_segments; self.gpx_segments.andand.size end
  def reimport_path; "/osm_trails/osmid-#{osmid}/reimport?returnto=/osm_trails/osmid-#{osmid}" end
  def guess_zone
    (osm_guideposts.first.andand.kctref || 'XX')[0..1]   # TODO: nějak zjemnit, teď bere jen první rozc.
  end
      
  private
  def self.color2barva(color)
    case color
      when 'red' then 'červená'
      when 'blue' then 'modrá'
      when 'green' then 'zelená'
      when 'yellow' then 'žlutá'
      when 'black' then 'černá'
      when 'orange' then 'oranžová'
      else raise "Neznámá anglická barva #{color}"
    end
  end
  
  def self.osmcsymbol2typ(osmcsymbol)
      if osmcsymbol.match(/^[^:]*:([^:]*):(red|blue|green|yellow|black|orange)_([^:]*)$/)
        background, symbol = $1,$3
      elsif osmcsymbol.match(/^(red|blue|green|yellow|black|orange)_(.*)$/) # když začátek chybí
        background, symbol = 'white',$2   
      elsif osmcsymbol=='none'
        background = 'none'
      elsif osmcsymbol.match(/^green:green_frame|red:white::/) # i.e. red:white::D1:red
        background = 'error'
      else
        background = 'error'
#         raise "Unrecognized osmcsymbol #{osmcsymbol}"
      end
      case background
        when 'white' then # yellow:white:L:yellow
          case symbol.strip
            when 'bar' then 'pěší'
            when 'turned_T','triangle','L','bowl' then 'odbočka'
            when 'backslash','slash','backlash' then 'naučná'
            when 'dot' then 'jezdecká'
            when 'corner' then 'místní'
            when 'pointer' then 'zvláštní' 
            when 'wheelchair','wheel' then 'vozíčkářská'
            when 'diamond','x', 'rectangle','backlash' then 'chybná' # časem opravovat na pravděpodobné
            else 'zvláštní' # raise "Unrecognized shape part '#{symbol.strip}' of osmcsymbol '#{osmcsymbol}'"
          end
        when 'yellow' then 'cyklistická'
        when 'orange' then 'lyžařská'
        when 'none' then 'neznačená'
        when 'error' then 'chybná'
        else 'chybná' # raise 'Unrecognized osmcsymbol '+ osmcsymbol
      end
  end
  
  def has_priority_osm_guideposts?
    self.osm_guideposts.each{|og| return true if og.is_priority_osm_guidepost? }
    return false
  end
  
  def self.SKIP_IDS 
    [1242417, 104838, 2434722, 3300718, 2869055, 2411711, 77738, 2010361, 3083510, 389137, 2153741, 3483133, 1556693, 1777623, 123925, 109682, 3200698, 105328, 1503231, 105165, 79437, 192031, 77061, 29133, 12295, 951694, 4621732, 166319, 109645, 1620840, 2083468, 1191538, 396110, 1830702, 2584352, 11019493, 9334311, 13161384 ] + # german, polish and slovak trails touching CZ
    [1698239,  569710, 569713, 7700604 ] + # slovak trails in CZ 
    [1017409, 1754959, 1819252, 188265, 165711, 188265, 156482, 106694, 69257, 11083, 3924352, 1191538 ] + # European long distance trail
    [10656261] # + # via ferrata
    #[13775089, 1981340, 12183420] # problematic large - investigate
    # guideposts  [244726260]
  end

  # symbol pouzivany k oznaceni osm tras v textovém výstupu obecně
  def self.symbol; '↪' end # '⤳'
  
end

